from setuptools import setup

setup(name='healthinsurance',
version='0.1.2',
description='Project Package',
url='https://www.linkedin.com/in/brunovn/',
author='bruno vinicius nonato',
author_email='brunovinicius154@gmail.com',
license='MIT',
packages=['healthinsurance'],
zip_safe=False)


